{
  Vampyre Imaging Library
  by Marek Mauder
  http://imaginglib.sourceforge.net

  The contents of this file are used with permission, subject to the Mozilla
  Public License Version 1.1 (the "License"); you may not use this file except
  in compliance with the License. You may obtain a copy of the License at
  http://www.mozilla.org/MPL/MPL-1.1.html

  Software distributed under the License is distributed on an "AS IS" basis,
  WITHOUT WARRANTY OF ANY KIND, either express or implied. See the License for
  the specific language governing rights and limitations under the License.

  Alternatively, the contents of this file may be used under the terms of the
  GNU Lesser General Public License (the  "LGPL License"), in which case the
  provisions of the LGPL License are applicable instead of those above.
  If you wish to allow use of your version of this file only under the terms
  of the LGPL License and not to allow others to use your version of this file
  under the MPL, indicate your decision by deleting  the provisions above and
  replace  them with the notice and other provisions required by the LGPL
  License.  If you do not delete the provisions above, a recipient may use
  your version of this file under either the MPL or the LGPL License.

  For more information about the LGPL: http://www.gnu.org/copyleft/lesser.html
}

{ This unit contains image format loader for WebP Codec images known as WEBP.}
unit ImagingWebP;

{$I ImagingOptions.inc}

interface

uses
  Classes, SysUtils, ImagingTypes, Imaging, ImagingIO, PasWebP;

type
  { Class for loading and saving WebP codec images.}
  TWEBPFileFormat = class(TImageFileFormat)
  protected
    procedure Define; override;
    function LoadData(Handle: TImagingHandle; var Images: TDynImageDataArray;
      OnlyFirstLevel: Boolean): Boolean; override;
    function SaveData(Handle: TImagingHandle; const Images: TDynImageDataArray;
      Index: LongInt): Boolean; override;
    procedure ConvertToSupported(var Image: TImageData;
      const Info: TImageFormatInfo); override;
  public
    function TestFormat(Handle: TImagingHandle): Boolean; override;
end;

implementation

const
  SWEBPFormatName = 'WebP Codec Image by Google';
  SWEBPMasks      = '*.webp';
  WebPSupportedFormats: TImageFormats = [ifR8G8B8, ifA8R8G8B8];

type
  TWEBPHeader = packed record
    RiffTag:  array [0..3] of AnsiChar;
    DataSize: Integer;
    WebPType: array [0..3] of AnsiChar;
    Codec:    array [0..3] of AnsiChar;
    CompSize: Integer;
  end;

procedure TWEBPFileFormat.Define;
begin
  inherited;
  FName := SWEBPFormatName;
  FFeatures := [ffLoad, ffSave];
  FSupportedFormats := WebPSupportedFormats;

  AddMasks(SWEBPMasks);
end;

function TWEBPFileFormat.LoadData(Handle: TImagingHandle;
  var Images: TDynImageDataArray; OnlyFirstLevel: Boolean): Boolean;
var
  Header: TWEBPHeader;
  Features: WebPBitstreamFeatures;
  Data: TBytes;
begin
  Result := False;
  with GetIO do
  begin
    Read(Handle, @Header, SizeOf(TWEBPHeader));
    Seek(Handle, -SizeOf(TWEBPHeader), smFromCurrent); // get back
    SetLength(Data, Header.DataSize + 8); // riff and size
    Read(Handle, @Data[0], Length(Data)); // read RAW

    WebPGetFeatures(@Data[0], Length(Data), @Features);
    if Features.Has_Animation = 0 then begin // Single image
      SetLength(Images, 1);
      if Features.Has_Alpha = 1 then begin
        InitImage(Images[0]);
        NewImage(Features.Width, Features.Height, ifA8R8G8B8, Images[0]);
        WebPDecodeRGBAInto(@Data[0], Length(Data), Images[0].Bits,
          Images[0].Size, Features.Width * 4);
      end else begin
        NewImage(Features.Width, Features.Height, ifR8G8B8, Images[0]);
        WebPDecodeRGBInto(@Data[0], Length(Data), Images[0].Bits,
          Images[0].Size, Features.Width * 3);
      end;
      Result := True;
    end; // not supported
    SetLength(Data, 0); // Free mem
  end;
end;

function TWEBPFileFormat.SaveData(Handle: TImagingHandle; const Images: TDynImageDataArray;
    Index: LongInt): Boolean;
var
  Config: WebPConfig;
  Pic: WebPPicture;
  Writer: WebPMemoryWriter;
  OK: Integer;
begin
  Result := False;
  with GetIO do
  begin
    if Length(Images) <> 1 then Exit;

    WebPConfigInit(@Config);
    WebPConfigLosslessPreset(@Config, 6); // Lossless - max
    Config.Quality := 100;

    if (WebPPictureInit(@Pic) = 0) then Exit;
    Pic.Width := Images[0].Width;
    Pic.Height := Images[0].Height;
    if (WebPPictureAlloc(@Pic) = 0) then Exit;
    if Images[0].Format = ifA8R8G8B8 then
      WebPPictureImportRGBA(@Pic, Images[0].Bits, Images[0].Width * 4)
    else
      WebPPictureImportRGB(@Pic, Images[0].Bits, Images[0].Width * 3);
    WebPMemoryWriterInit(@Writer);
    Pic.Writer := @WebPMemoryWrite;
    Pic.Custom_Ptr := @Writer;

    OK := WebPEncode(@Config, @Pic);
    Write(Handle, Writer.Mem, Writer.Size);
    WebPPictureFree(@Pic);
    WebPMemoryWriterClear(@Writer);

    if OK <> 0 then
      Result := True;
  end;
end;

procedure TWEBPFileFormat.ConvertToSupported(var Image: TImageData;
  const Info: TImageFormatInfo);
var
  ConvFormat: TImageFormat;
begin
  if Info.HasAlphaChannel then
    // Convert images with alpha channel to A8R8G8B8
    ConvFormat := ifA8R8G8B8
  else
    // Convert all other formats to R8G8B8
    ConvFormat := ifR8G8B8;

  ConvertImage(Image, ConvFormat);
end;

function TWEBPFileFormat.TestFormat(Handle: TImagingHandle): Boolean;
var
  Hdr: TWEBPHeader;
  ReadCount: LongInt;
begin
  Result := False;
  if Handle <> nil then
  begin
    ReadCount := GetIO.Read(Handle, @Hdr, SizeOf(TWEBPHeader));
    if (Hdr.RiffTag = 'RIFF') and (Hdr.WebPType = 'WEBP') then
      Result := True;
    GetIO.Seek(Handle, -ReadCount, smFromCurrent);
  end;
end;

initialization
  RegisterImageFileFormat(TWEBPFileFormat);

end.

